"""
This scripts generates the listing of files that the CLI will download.

It does so by partitioning a list of ManifestEntry's into parts. See the
`manifests` dict below; the keys are the names of each part of the dataset.

This should only be used internally by the Eng Team of the EgoExo dataset.
"""

import json
import os
from collections import defaultdict

from ego4d.internal.download.manifest import (
    manifest_dumps,
    ManifestEntry,
    PathSpecification,
)

from ego4d.internal.s3 import S3Downloader
from iopath.common.file_io import PathManager
from iopath.common.s3 import S3PathHandler
from tqdm.auto import tqdm

pathmgr = PathManager()  # for downloading files
pathmgr.register_handler(S3PathHandler(profile="default"))

dev_release = False

if dev_release:
    ver_name = "dev"
    base_release_dir = "egoexo/releases/dev/"
    release_dir = f"s3://ego4d-consortium-sharing/{base_release_dir}"
    internal_release_dir = release_dir
else:
    ver_name = "v2"
    base_release_dir = f"egoexo-public/{ver_name}/"
    release_dir = f"s3://ego4d-consortium-sharing/{base_release_dir}"
    internal_release_dir = (
        f"s3://ego4d-consortium-sharing/egoexo/releases/public_internal/{ver_name}/"
    )


manifests = {
    "metadata": [],
    "annotations": [],
    "takes": [],
    "captures": [],
    "take_trajectory": [],
    "take_eye_gaze": [],
    "take_point_cloud": [],
    "take_vrs": [],
    "take_vrs_noimagestream": [],
    "capture_trajectory": [],
    "capture_eye_gaze": [],
    "capture_point_cloud": [],
    "downscaled_takes/448": [],
    "features/omnivore_video": [],
    "features/maws_clip_2b": [],
    "ego_pose_pseudo_gt": [],
    "expert_commentary": [],
    "take_transcription": [],
    "take_audio": [],
}
if dev_release:
    manifests["takes_dropped"] = []

egoexo = {
    "released_takes": os.path.join(internal_release_dir, "released_takes.json"),
    "takes": os.path.join(internal_release_dir, "takes.json"),
    "takes_dropped": os.path.join(internal_release_dir, "_takes_dropped.json"),
    "captures": os.path.join(internal_release_dir, "captures.json"),
    "physical_setting": os.path.join(internal_release_dir, "physical_setting.json"),
    "participants": os.path.join(internal_release_dir, "participants.json"),
    "visual_objects": os.path.join(internal_release_dir, "visual_objects.json"),
    "metadata": os.path.join(internal_release_dir, "metadata.json"),
}

downloader = S3Downloader("default")
print("Downloading metadata")
for k, out_path in tqdm(egoexo.items()):
    s3_path = os.path.join(release_dir, f"{k}.json")
    if not pathmgr.exists(s3_path):
        print(f"WARN: {s3_path} does not exist")
        continue
    # print(s3_path, out_path)
    paths = [
        PathSpecification(
            s3_path,
            f"{k}.json",
        )
    ]
    manifests["metadata"].append(
        ManifestEntry(
            uid=k,
            paths=paths,
        )
    )

for k, v in egoexo.items():
    if pathmgr.exists(v):
        egoexo[k] = json.load(pathmgr.open(v))
    else:
        egoexo[k] = []

if len(egoexo["released_takes"]) == 0:
    egoexo["released_takes"] = [x["take_uid"] for x in egoexo["takes"]]

s3_buckets = set()
for c in tqdm(egoexo["captures"]):
    s3_bucket = c["_s3_root_dir"].split("/")[2]
    s3_buckets.add(s3_bucket)

take_by_take_uid = {x["take_uid"]: x for x in egoexo["takes"] + egoexo["takes_dropped"]}
take_name_to_uid = {
    t["take_name"]: t["take_uid"] for t in (egoexo["takes"] + egoexo["takes_dropped"])
}

capture_cam_id_to_is_ego = {
    (c["capture_uid"], cam["cam_id"]): cam["is_ego"]
    for c in egoexo["captures"]
    for cam in c["cameras"]
}

# split_file_path = os.path.join(release_dir, "annotations/splits.json")
split_file_path = "s3://ego4d-consortium-sharing/egoexo-public/v2/annotations/splits.json"
split_data = json.load(pathmgr.open(split_file_path))
take_uid_to_splits = {k: [v] for k, v in split_data["take_uid_to_split"].items()}

capture_uid_to_splits = defaultdict(set)
for take_uid, split in split_data["take_uid_to_split"].items():
    if take_uid not in take_by_take_uid:
        print("Missing", take_uid)
        continue
    capture_uid = take_by_take_uid[take_uid]["capture"]["capture_uid"]
    capture_uid_to_splits[capture_uid].add(split.lower())

take_uid_to_benchmarks = defaultdict(list)
take_uid_to_benchmarks.update(split_data.get("take_uid_to_benchmark", {}))

s3_buckets = {
    x["university_id"]
    for x in egoexo["takes"]
    if x["take_uid"] in egoexo["released_takes"]
}

manifests["downscaled_takes/448"] = []
if "downscaled_takes/448" in manifests:
    by_take = defaultdict(list)
    for bucket in s3_buckets:
        ds_base_dir = os.path.join(
            f"s3://{bucket}/{base_release_dir}", "downscaled_takes/448/"
        )
        print(ds_base_dir)
        for bn, path in downloader.ls(ds_base_dir, recursive=True):
            take_name = path.split("downscaled_takes/")[1].split("/")[1]
            if take_name not in take_name_to_uid:
                continue
            take_uid = take_name_to_uid[take_name]
            if take_uid not in egoexo["released_takes"]:
                continue
            by_take[take_uid].append(
                PathSpecification(
                    source_path=path,
                    relative_path=f"takes/{take_name}/frame_aligned_videos/downscaled/448/{bn}",
                    file_type="mp4",
                    views=None,  # TODO
                )
            )

    for take_uid, paths in by_take.items():
        manifests["downscaled_takes/448"].append(
            ManifestEntry(
                uid=take_uid,
                paths=paths,
                benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                splits=take_uid_to_splits.get(take_uid, None),
            )
        )

path = None
take_uid = None
take_name = None
by_take_and_dir = {}
for bucket in tqdm(s3_buckets):
    base_dir = os.path.join(f"s3://{bucket}/{base_release_dir}", "takes/")
    for bn, path in downloader.ls(base_dir, recursive=True):
        take_name = path.split("takes/")[1].split("/")[0]
        if take_name not in take_name_to_uid:
            continue
        take_uid = take_name_to_uid[take_name]
        if take_uid not in egoexo["released_takes"]:
            continue
        dirname = os.path.basename(os.path.dirname(path))
        key = (take_uid, dirname)
        if key not in by_take_and_dir:
            by_take_and_dir[key] = []
        by_take_and_dir[key].append((bn, path))


for manifest_key in ["takes", "takes_dropped"]:
    if manifest_key not in manifests:
        continue

    for t in tqdm(egoexo[manifest_key]):
        take_uid = t["take_uid"]
        if take_uid not in egoexo["released_takes"]:
            continue

        root_dir = t["root_dir"]
        paths = []
        for streams in t["frame_aligned_videos"].values():
            for vid in streams.values():
                if vid["_s3_path"] is None:
                    continue
                uid = vid["clip_uid"]
                vid_k = (t["capture_uid"], vid.get("cam_id"))
                is_ego = capture_cam_id_to_is_ego.get(vid_k)
                views = None
                if is_ego is not None:
                    views = ["ego"] if is_ego else ["exo"]
                paths.append(
                    PathSpecification(
                        source_path=vid["_s3_path"],
                        relative_path=os.path.join(root_dir, vid["relative_path"]),
                        views=views,
                        universities=[t["university_name"]],
                        file_type="mp4",
                    )
                )

        # add vrs path (no rgb stream)
        if t["has_trimmed_vrs"]:
            for d in ["trajectory", "eye_gaze"]:
                d_paths = []
                pc_paths = []
                key = (take_uid, d)
                if key not in by_take_and_dir:
                    print("Skipping", key)
                    continue
                for bn, p in by_take_and_dir[key]:
                    ext = os.path.splitext(bn)[-1]
                    if "semidense" in bn:
                        pc_paths.append(
                            PathSpecification(
                                source_path=p,
                                relative_path=os.path.join(root_dir, d, bn),
                                views=None,
                                universities=[t["university_name"]],
                                file_type=ext[1:],
                            )
                        )
                    else:
                        d_paths.append(
                            PathSpecification(
                                source_path=p,
                                relative_path=os.path.join(root_dir, d, bn),
                                views=None,
                                universities=[t["university_name"]],
                                file_type=ext[1:],
                            )
                        )

                if len(pc_paths) > 0:
                    manifests["take_point_cloud"].append(
                        ManifestEntry(
                            uid=take_uid,
                            paths=pc_paths,
                            benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                            splits=take_uid_to_splits.get(take_uid, None),
                        )
                    )

                if d != "eye_gaze":
                    assert d == "trajectory"
                    manifests["take_trajectory"].append(
                        ManifestEntry(
                            uid=take_uid,
                            paths=d_paths,
                            benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                            splits=take_uid_to_splits.get(take_uid, None),
                        )
                    )
                else:
                    assert d == "eye_gaze"
                    manifests["take_eye_gaze"].append(
                        ManifestEntry(
                            uid=take_uid,
                            paths=d_paths,
                            benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                            splits=take_uid_to_splits.get(take_uid, None),
                        )
                    )

            sp = t["_vrs_s3_path"]
            if sp is not None:
                manifests["take_vrs_noimagestream"].append(
                    ManifestEntry(
                        uid=take_uid,
                        paths=[
                            PathSpecification(
                                source_path=sp,
                                relative_path=os.path.join(
                                    root_dir, t["vrs_relative_path"]
                                ),
                                views=["ego"],
                                universities=[t["university_name"]],
                                file_type="vrs",
                            )
                        ],
                        benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                        splits=take_uid_to_splits.get(take_uid, None),
                    )
                )
            
            sp_all_streams = t["_vrs_all_streams_s3_path"]
            if sp_all_streams is not None:
                manifests["take_vrs"].append(
                    ManifestEntry(
                        uid=take_uid,
                        paths=[
                            PathSpecification(
                                source_path=sp_all_streams,
                                relative_path=os.path.join(
                                    root_dir, t["vrs_all_streams_relative_path"]
                                ),
                                views=["ego"],
                                universities=[t["university_name"]],
                                file_type="vrs",
                            )
                        ],
                        benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                        splits=take_uid_to_splits.get(take_uid, None),
                    )
                )

        manifests[manifest_key].append(
            ManifestEntry(
                uid=take_uid,
                paths=paths,
                benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                splits=take_uid_to_splits.get(take_uid, None),
            )
        )

if "take_transcription" in manifests:
    assert "take_audio" in manifests
    for take in egoexo["takes"]:
        take_uid = take["take_uid"]
        root_dir = take["root_dir"]
        audio_root_dir = os.path.join(root_dir, "audio")
        key = (take_uid, "audio")
        if key not in by_take_and_dir:
            continue
        
        audio_paths = []
        transcription_paths = []
        for bn, p in by_take_and_dir[key]:
            if bn.endswith(".json"):
                transcription_paths += [
                    PathSpecification(
                        source_path=p,
                        relative_path=os.path.join(root_dir, "audio", bn),
                        views=None,
                        universities=None,
                    )
                ]
            else:
                assert bn.endswith(".wav")
                audio_paths += [
                    PathSpecification(
                        source_path=p,
                        relative_path=os.path.join(root_dir, "audio", bn),
                        views=None,
                        universities=None,
                    )
                ]

        manifests["take_audio"].append(
            ManifestEntry(
                uid=take_uid,
                paths=audio_paths,
                benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                splits=take_uid_to_splits.get(take_uid, None),
            )
        )
        manifests["take_transcription"].append(
            ManifestEntry(
                uid=take_uid,
                paths=transcription_paths,
                benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                splits=take_uid_to_splits.get(take_uid, None),
            )
        )

for c in tqdm(egoexo["captures"]):
    s3_bucket = c["_s3_root_dir"].split("/")[2]
    root_dir = c["root_dir"]
    traj_fp = c["_trajectory_s3_dir"]
    post_fp = c["_postsurvery_s3_path"]
    eye_gaze_fp = c["_gaze_s3_dir"]
    timesync_fp = c["_timesync_s3_path"]
    ts_dir = os.path.join(c["_s3_root_dir"], "timesync")
    traj_files = downloader.ls(traj_fp + "/", max_keys=1000) if traj_fp else []
    eye_gaze_files = (
        downloader.ls(eye_gaze_fp + "/", max_keys=1000) if eye_gaze_fp else []
    )
    ts_files = downloader.ls(ts_dir + "/", max_keys=1000)
    universities = [c["university_name"]]

    traj_paths = []
    point_cloud_paths = []
    for bn, s3_path in traj_files:
        ext = os.path.splitext(bn)[-1]
        assert ext.startswith(".")
        if "semidense" in bn:
            point_cloud_paths.append(
                PathSpecification(
                    source_path=s3_path,
                    relative_path=os.path.join(root_dir, "trajectory", bn),
                    universities=universities,
                    file_type=ext[1:],
                    views=None,
                )
            )
        else:
            traj_paths.append(
                PathSpecification(
                    source_path=s3_path,
                    relative_path=os.path.join(root_dir, "trajectory", bn),
                    universities=universities,
                    file_type=ext[1:],
                    views=None,
                )
            )

    capture_uid = c["capture_uid"]
    manifests["capture_trajectory"].append(
        ManifestEntry(
            uid=capture_uid,
            paths=traj_paths,
            benchmarks=take_uid_to_benchmarks[capture_uid],
            splits=capture_uid_to_splits[capture_uid],
        )
    )
    manifests["capture_point_cloud"].append(
        ManifestEntry(
            uid=capture_uid,
            paths=point_cloud_paths,
            benchmarks=take_uid_to_benchmarks[capture_uid],
            splits=capture_uid_to_splits[capture_uid],
        )
    )
    eye_gaze_paths = []
    for bn, s3_path in eye_gaze_files:
        ext = os.path.splitext(bn)[-1]
        assert ext.startswith(".")
        eye_gaze_paths.append(
            PathSpecification(
                source_path=s3_path,
                relative_path=os.path.join(root_dir, "eye_gaze", bn),
                universities=universities,
                file_type=ext[1:],
                views=None,
            )
        )
    manifests["capture_eye_gaze"].append(
        ManifestEntry(
            uid=capture_uid,
            paths=eye_gaze_paths,
            benchmarks=take_uid_to_benchmarks[capture_uid],
            splits=capture_uid_to_splits[capture_uid],
        )
    )

    capture_paths = []
    if post_fp is not None:
        capture_paths.append(
            PathSpecification(
                source_path=post_fp,
                relative_path=os.path.join(root_dir, "post_surveys.csv"),
                universities=universities,
                file_type="csv",
                views=None,
            )
        )

    if timesync_fp is not None:
        capture_paths.append(
            PathSpecification(
                source_path=timesync_fp,
                relative_path=os.path.join(root_dir, "timesync.csv"),
                universities=universities,
                file_type="csv",
                views=None,
            )
        )
    for bn, s3_path in ts_files:
        ext = os.path.splitext(bn)[-1]
        assert ext.startswith(".")
        capture_paths.append(
            PathSpecification(
                source_path=s3_path,
                relative_path=os.path.join(root_dir, "timesync", f"{bn}"),
                universities=universities,
                file_type=ext[1:],
                views=None,
            )
        )

    manifests["captures"].append(
        ManifestEntry(
            uid=c["capture_uid"],
            paths=capture_paths,
            benchmarks=take_uid_to_benchmarks[capture_uid],
            splits=capture_uid_to_splits[capture_uid],
        )
    )


if "annotations" in manifests:
    manifests["annotations"] = []
    annotations = downloader.ls(os.path.join(release_dir, "annotations/"))
    for bn, s3_path in annotations:
        if len(bn) == 0:
            continue
        if bn == "manifest.json":
            continue

        print(bn)
        benchmarks = []
        benchmark_name = "_".join(bn.split("_")[0:-1])
        benchmarks = [benchmark_name] if len(benchmark_name) > 0 else None
        if "proficiency" in benchmark_name:
            benchmarks.append("proficiency")

        manifests["annotations"].append(
            ManifestEntry(
                uid="_".join(bn.split("_")[:-1]),
                paths=[
                    PathSpecification(
                        source_path=s3_path,
                        relative_path=f"annotations/{bn}",
                    )
                ],
                benchmarks=benchmarks,
            )
        )

    egopose_part_subdirs = {
        "annotations": {
            "subdirs": {
                "body": ["annotation"],
                "hand": ["annotation"],
                "camera_pose": [""],
            },
            "take_uid_as_key": False,
        },
        "ego_pose_pseudo_gt": {
            "subdirs": {
                "body": ["automatic"],
                "hand": ["automatic"],
                "camera_pose": [],
            },
            "take_uid_as_key": True,
        },
    }
    if not dev_release:
        egopose_base_dir = os.path.join(release_dir, "annotations/ego_pose/")
        for split in ["train", "val"]:
            for part in ["body", "hand", "camera_pose"]:
                subdir = os.path.join(egopose_base_dir, split, part)
                print(subdir)

                annotation_files = []

                # go over each potential manifest --part
                for manifest_key, metadata in egopose_part_subdirs.items():
                    for ann_dir in metadata["subdirs"][part]:
                        dst_relative_subdir = os.path.join(
                            "annotations/ego_pose/", split, part, ann_dir
                        )
                        abs_subdir = os.path.join(subdir, ann_dir)

                        if not abs_subdir.endswith("/"):
                            abs_subdir += "/"
                        ann_files = downloader.ls(abs_subdir)
                        for bn, s3_path in ann_files:
                            if len(bn) == 0:
                                continue
                            if bn == "manifest.json":
                                continue
                            take_uid = os.path.splitext(bn)[0]
                            uid = take_uid
                            splits = take_uid_to_splits.get(take_uid, [])
                            assert len(splits) == 1 and splits[0] == split, "split diff: {splits} vs {split}"
                            benchmarks = ["egopose", "ego_pose"]
                            if part in ["body", "hand"]:
                                benchmarks += [
                                    f"{part}_pose",
                                    f"ego_{part}_pose",
                                    f"{part}pose",
                                    f"ego{part}pose",
                                    f"ego_{part}pose",
                                ]
                            else:
                                assert part == "camera_pose"
                                for body_part in ["body", "hand"]:
                                    benchmarks += [
                                        f"{body_part}_pose",
                                        f"ego_{body_part}_pose",
                                        f"{body_part}pose",
                                        f"ego{body_part}pose",
                                        f"ego_{body_part}pose",
                                        f"camera_pose",
                                    ]

                            manifests[manifest_key].append(
                                ManifestEntry(
                                    uid=uid,
                                    benchmarks=benchmarks,
                                    splits=splits,
                                    paths=[
                                        PathSpecification(
                                            source_path=s3_path,
                                            relative_path=f"{dst_relative_subdir}/{bn}",
                                        )
                                    ],
                                )
                          )
    # egopose_anns = [
    #     x for x in manifests["annotations"]
    #     if x.benchmarks is not None and "egopose" in x.benchmarks
    #     # if x.benchmarks is not None and "body_pose" in x.benchmarks
    # ]

if "expert_commentary" in manifests:
    manifests["expert_commentary"] = []
    ec_base_dir = os.path.join(
        release_dir, "annotations/expert_commentary/"
    )
    fs = downloader.ls(ec_base_dir, recursive=True)
    for _, f in tqdm(fs):
        base_dir = f[len(ec_base_dir):]
        take_name, expert_name = base_dir.split("/")[0:2]
        if take_name not in take_name_to_uid:
            continue
        take_uid = take_name_to_uid[take_name]
        uid = f"{take_uid}_{expert_name}"
        splits = take_uid_to_splits[take_uid]

        manifests["expert_commentary"].append(
            ManifestEntry(
                uid=uid,
                benchmarks=["expert_commentary"],
                splits=splits,
                paths=[
                    PathSpecification(
                        source_path=f,
                        relative_path=os.path.join("annotations/expert_commentary/", base_dir),
                    ),
                ],
            )
        )

if "features/omnivore_video" in manifests:
    for feature_name in ["omnivore_video", "maws_clip_2b"]:
        feature_files = downloader.ls(
            os.path.join(release_dir, "features/", feature_name + "/")
        )
        by_take = defaultdict(list)
        for file_name, path in feature_files:
            if file_name.endswith("yaml"):
                manifests[f"features/{feature_name}"].append(
                    ManifestEntry(
                        uid="omnivore_video_config",
                        paths=[
                            PathSpecification(
                                source_path=path,
                                relative_path=f"features/{feature_name}/{file_name}",
                            )
                        ],
                    )
                )
                continue

            if file_name == "" or file_name == "manifest.json":
                continue
            take_uid, cam_id, stream_id = file_name.split("_")
            if take_uid not in egoexo["released_takes"]:
                continue

            t = take_by_take_uid[take_uid]
            vid_k = (t["capture_uid"], vid.get("cam_id"))
            is_ego = capture_cam_id_to_is_ego.get(vid_k)
            views = None
            if is_ego is not None:
                views = ["ego"] if is_ego else ["exo"]
            by_take[take_uid].append(
                PathSpecification(
                    source_path=path,
                    relative_path=f"features/{feature_name}/{file_name}",
                    views=views,
                    universities=[t["university_name"]],
                    file_type="pt",
                )
            )

        for take_uid, paths in by_take.items():
            manifests[f"features/{feature_name}"].append(
                ManifestEntry(
                    uid=take_uid,
                    paths=paths,
                    benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                    splits=take_uid_to_splits.get(take_uid, None),
                )
            )


all_bs = set()
for k, v in manifests.items():
    out_dir = os.path.join(release_dir, f"{k}/")
    manifest_file = os.path.join(out_dir, "manifest.json")
    for m in v:
        for b in m.benchmarks or []:
            all_bs.add(b)

    print(k, len(v))
    print(f"will output to: {out_dir}")
    print(f"{manifest_file}")
    pathmgr.mkdirs(out_dir, exist_ok=True)
    pathmgr.rm(manifest_file)
    with pathmgr.open(manifest_file, "w") as out_f:
        json_data = manifest_dumps(v)
        out_f.write(json_data)



import argparse
import os
import sys
import threading
import traceback
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, List, Optional, Tuple, TypeVar

from ego4d.cli.progressbar import DownloadProgressBar
from ego4d.cli.universities import UNIV_TO_BUCKET

from ego4d.internal.download.manifest import (
    manifest_loads,
    ManifestEntry,
    PathSpecification,
)
from ego4d.internal.s3 import S3Downloader
from iopath.common.file_io import PathManager
from iopath.common.s3 import S3PathHandler
from tqdm.auto import tqdm

T = TypeVar("T")
U = TypeVar("U")


def _s3_path_join(a: str, b: str) -> str:
    assert a.startswith("s3://")
    return os.path.join(a, b).replace("\\", "/")  # hacky, but it works


def _path_ok(p: PathSpecification, args) -> bool:
    ok = True
    if (
        args.views is not None
        and p.views is not None
        and len(set(p.views) & args.views) == 0
    ):
        ok = False
    if (
        args.universities is not None
        and p.universities is not None
        and len(set(p.universities) & args.universities) == 0
    ):
        ok = False
    return ok


def _manifest_ok(m: ManifestEntry, args) -> bool:
    ok = True
    if (
        args.splits is not None
        and m.splits is not None
        and len(set(m.splits) & args.splits) == 0
    ):
        ok = False
    if (
        args.benchmarks is not None
        and m.benchmarks is not None
        and len(set(m.benchmarks) & args.benchmarks) == 0
    ):
        ok = False
    if args.uids is not None and m.uid is not None and len({m.uid} & args.uids) == 0:
        ok = False
    return ok


def map_all(
    values: List[T],
    map_fn: Callable[[Optional[S3Downloader], T], Tuple[U, T, Optional[str]]],
    num_workers: int,
    s3_profile: Optional[str],
    needs_downloader: bool,
    progress_on_bytes: bool,
    total_bytes: Optional[int],
) -> Tuple[List[Tuple[T, U]], List[Tuple[T, str]]]:
    # NOTE: boto3 is not thread safe
    thread_data = threading.local()
    callback = None

    def initializer():
        if needs_downloader:
            thread_data.downloader = S3Downloader(
                s3_profile,
                num_workers=num_workers,
                callback=callback,
            )
        else:
            thread_data.downloader = None

    def wrap_map(x):
        return map_fn(thread_data.downloader, x)

    failures = []
    ret = []
    iterator_wrapper = tqdm
    if progress_on_bytes:
        progress = DownloadProgressBar(total_bytes)
        iterator_wrapper = lambda x, total: x
        callback = progress.update if progress is not None else None

    with ThreadPoolExecutor(max_workers=num_workers, initializer=initializer) as pool:
        for value, key, err in iterator_wrapper(
            pool.map(wrap_map, values), total=len(values)
        ):
            if value is None:
                failures.append((key, err))
            else:
                ret.append((key, value))
    return ret, failures


def main(args):
    base_dir = args.base_dir
    release_name = args.release
    parts = set(args.parts)
    out_dir = args.out_dir
    num_workers = args.num_workers
    s3_profile = args.s3_profile
    yes = args.yes
    args.uids = set(args.uids) if args.uids is not None else None
    args.benchmarks = set(args.benchmarks) if args.benchmarks is not None else None
    args.splits = set(args.splits) if args.splits is not None else None
    args.views = set(args.views) if args.views is not None else None
    args.universities = (
        set(args.universities) if args.universities is not None else None
    )

    assert (
        release_name is not None
    ), """You must provide a release name via `--release`.
If you meant to download the public release, please use the script `ego4d/egoexo/download/cli.py` instead.
"""

    # TODO: remove iopath dependency
    pathmgr = PathManager()
    pathmgr.register_handler(S3PathHandler(profile=s3_profile))

    def check_file(
        downloader: Optional[S3Downloader],
        path_expected_size: Tuple[PathSpecification, int],
    ) -> Tuple[Optional[int], Tuple[PathSpecification, int], Optional[str]]:
        assert downloader is None
        path, expected_size = path_expected_size
        out_path = os.path.join(out_dir, path.relative_path)
        try:
            if not os.path.exists(out_path):
                return 0, path_expected_size, None
            size = os.path.getsize(out_path)
            ok = size == expected_size
            return size if ok else 0, path_expected_size, None
        except Exception:
            return None, path_expected_size, traceback.format_exc()

    def get_size(
        downloader: Optional[S3Downloader], path: PathSpecification
    ) -> Tuple[Optional[int], PathSpecification, Optional[str]]:
        assert downloader is not None
        try:
            if not path.source_path.startswith("s3://"):
                return None, path, "path is not an S3 path"
            return downloader.obj(path.source_path).content_length, path, None
        except Exception:
            return None, path, traceback.format_exc()

    def download(
        downloader: Optional[S3Downloader],
        path_size_pair: Tuple[PathSpecification, int],
    ) -> Tuple[Optional[int], Tuple[PathSpecification, int], Optional[str]]:
        assert downloader is not None
        path, _ = path_size_pair
        try:
            out_path = os.path.join(out_dir, path.relative_path)
            downloader.copy(path.source_path, out_path)
            size, _, err = check_file(None, path_size_pair)
            return size, path_size_pair, err
        except Exception:
            return None, path_size_pair, traceback.format_exc()

    os.makedirs(out_dir, exist_ok=True)
    assert os.path.isdir(out_dir), f"output dir {out_dir} is not a directory"

    release_dir = _s3_path_join(base_dir, release_name)
    if not release_dir.endswith("/"):
        release_dir += "/"

    num_paths = 0
    all_paths = []
    part_errs = []
    for part in parts:
        manifest_path = _s3_path_join(_s3_path_join(release_dir, part), "manifest.json")
        if not pathmgr.exists(manifest_path):
            part_errs.append(part)
            continue
        ms = manifest_loads(pathmgr.open(manifest_path).read())
        for m in ms:
            num_paths += len(m.paths)
            if not _manifest_ok(m, args):
                continue

            all_paths.extend([p for p in m.paths if _path_ok(p, args)])

    if len(part_errs) == len(parts):
        print("ERROR: could not get manifests for all parts (nothing to download)")
        print(
            """NOTE: Did you configure your AWS client correctly? See: https://github.com/facebookresearch/Ego4d/issues/277

This could also be due to your internet connection.

If you are located in China, please try using a VPN. Please refer to these posts:
- https://github.com/facebookresearch/Ego4d/issues/223
- https://github.com/facebookresearch/Ego4d/issues/162
- https://github.com/facebookresearch/Ego4d/issues/284
"""
        )
        print("Quitting ...")
        sys.exit(1)
    elif len(part_errs) > 0:
        print(f"ERROR: some supplied parts ({part_errs}) do not exist")
        print("NOTE: did you spell the parts correctly?")
        print("Quitting ...")
        sys.exit(2)

    if num_paths != len(all_paths):
        print(f"Filtered {num_paths} -> {len(all_paths)} files")

    # TODO: pre-cache this such that it is faster
    print("Determining what to download ...")
    path_size_pairs, s3_stat_failures = map_all(
        all_paths,
        map_fn=get_size,
        num_workers=num_workers,
        s3_profile=s3_profile,
        needs_downloader=True,
        progress_on_bytes=False,
        total_bytes=None,
    )
    s3_zero_sizes = [path for path, size in path_size_pairs if size == 0]
    success_path_size_pairs = [
        (path, size) for path, size in path_size_pairs if size != 0
    ]
    all_s3_stat_failures = len(s3_zero_sizes) + len(s3_stat_failures)
    if all_s3_stat_failures > 0:
        bucket_failures = {
            path.source_path.split("s3://")[1].split("/")[0]
            for path, _ in s3_stat_failures
        }
        print(
            f"""WARN: failed to get stats for {all_s3_stat_failures} files, will skip. [zero_size={len(s3_zero_sizes)}, exceptions={all_s3_stat_failures}.

This could be due to your internet connection.

If you are located in China, please try using a VPN. Please refer to these posts:
- https://github.com/facebookresearch/Ego4d/issues/223
- https://github.com/facebookresearch/Ego4d/issues/162
- https://github.com/facebookresearch/Ego4d/issues/284

Additional debugging information:
- S3 Buckets failed: {bucket_failures}
"""
        )

    total_size_bytes = sum(x for _, x in path_size_pairs if x is not None)
    total_size_gib = total_size_bytes / 1024**3

    print("Checking current download status ...")
    curr_paths, curr_stats_failures = map_all(
        success_path_size_pairs,
        map_fn=check_file,  # pyre-ignore
        num_workers=num_workers,
        s3_profile=s3_profile,
        needs_downloader=False,
        progress_on_bytes=False,
        total_bytes=None,
    )
    # TODO: refactor such that we can tell the number of files to be updated
    existing_paths = {
        path_size_pair
        for path_size_pair, actual_size in curr_paths
        if actual_size is not None and actual_size > 0
    }
    existing_gib = sum(size / 1024**3 for _, size in existing_paths if size is not None)
    existing_len = len(existing_paths)
    progress_percent = existing_gib / total_size_gib if total_size_gib > 0 else 1.0

    print(
        f"Downloaded: {progress_percent:.3%} = {existing_gib:.3f}GiB / {total_size_gib:.3f}GiB ({existing_len} / {len(path_size_pairs)} files) downloaded"
    )
    if len(curr_stats_failures) > 0:
        print(f"WARN: {len(curr_stats_failures)} failed to stat")

    ps_to_dl = {
        path: size
        for path, size in success_path_size_pairs
        if (path, size) not in existing_paths
    }
    if len(ps_to_dl) == 0 and not args.force:
        print("Everything has been downloaded. Bye.")
        sys.exit(0)

    if args.force:
        print("Forcing everything to be re-downloaded ...")
        ps_to_dl = {path: size for path, size in success_path_size_pairs}

    assert all(x is not None for x in ps_to_dl.values())
    expected_gb = sum(x / 1024**3 for x in ps_to_dl.values() if x is not None)
    confirm = False
    if not yes:
        response = input(
            f"Expected size of downloaded files is "
            f"{expected_gb:.1f} GiB ({len(ps_to_dl)} files). "
            f"Do you want to start the download? ([y]/n): "
        )
        if response.lower() in ["yes", "y", ""]:
            confirm = True
        else:
            confirm = False
    else:
        confirm = True

    if not confirm:
        print("Aborting...")
        sys.exit(0)

    print("Preparing output directories ...")
    all_out_dirs = {
        os.path.join(out_dir, os.path.dirname(x.relative_path)) for x in all_paths
    }
    for x in tqdm(all_out_dirs):
        os.makedirs(x, exist_ok=True)

    if args.delete:
        print("Scanning for files to delete ...")
        files_that_exist = []
        for dirpath, _, filenames in os.walk(out_dir):
            files_that_exist.extend([os.path.join(dirpath, f) for f in filenames])

        files_that_exist = set(files_that_exist)
        files_to_delete = files_that_exist - {
            os.path.join(out_dir, x[0].relative_path) for x in success_path_size_pairs
        }
        print(
            f"Deleting: {len(files_to_delete)} files ({len(files_that_exist)} total files)"
        )
        breakpoint()
        for f in tqdm(files_to_delete):
            os.remove(f)

    print("Downloading ...")
    assert all(size is not None for size in ps_to_dl.values())
    paths_to_fetch = [
        (path, size) for path, size in ps_to_dl.items() if size is not None
    ]
    total_bytes_to_fetch = sum(size for _, size in paths_to_fetch)
    print(f"Fetching: {total_bytes_to_fetch/1024**3:.3f}GiB")
    paths_downloaded, dl_failures = map_all(
        paths_to_fetch,  # pyre-ignore
        map_fn=download,  # pyre-ignore
        num_workers=num_workers,
        s3_profile=s3_profile,
        needs_downloader=True,
        progress_on_bytes=True,
        total_bytes=total_bytes_to_fetch,
    )

    integrity_errs = [path for path, size in paths_downloaded if size == 0]
    if len(dl_failures) > 0:
        num_all_failures = len(dl_failures) + len(integrity_errs)
        print(
            f"WARN: failed to fetch {num_all_failures} files [integrity={len(integrity_errs)}, exceptions={len(dl_failures)}]"
        )
        print("Please retry the download (... returning with error code 2)")
        sys.exit(2)
    print(flush=True)


def create_arg_parse(script_name: str, base_dir: str, release_name: Optional[str]):
    parser = argparse.ArgumentParser(
        usage=f"""
    EgoExo downloader CLI

    Simple usage:
        {script_name} -o <out_dir>

    Advanced usage examples:
        - Download point clouds and annotations 
            {script_name} -o <out_dir> --parts annotations take_point_cloud -y
        - Download VRS files (with RGB stream) for a take
            {script_name} -o <out_dir> --parts take_vrs --uids <uid1>

"""
    )
    parser.add_argument(
        "-o",
        "--out_dir",
        type=str,
        default=None,
        help="Which folder to download the data to",
        required=True,
    )
    parser.add_argument(
        "--parts",
        type=str,
        nargs="+",
        default=[
            "metadata",
            "captures",
            "takes",
            "take_trajectory",
            "take_vrs_noimagestream",
            "annotations",
        ],
        help="""
What parts of the dataset to download, one of:
- metadata
- annotations
- takes
- take_vrs_noimagestream
- take_trajectory
- take_point_cloud
- take_vrs
- take_audio
- take_transcription
- captures
- capture_trajectory
- capture_eye_gaze
- downscaled_takes/448
- features/omnivore_video
- features/maws_clip_2b
- ego_pose_pseudo_gt
- expert_commentary

By default the following parts will be downloaded: {metadata, captures, takes, trajectory, annotations}.

Example usage: --parts annotations point_cloud eye_gaze
""",
    )
    parser.add_argument(
        "--uids",
        type=str,
        nargs="+",
        default=None,
        help="what uids to filter for takes or captures",
    )
    parser.add_argument(
        "--benchmarks",
        type=str,
        nargs="+",
        default=None,
        help="""
Data relating to particular benchmarks will be only included. Valid benchmark
values are: keystep, proficiency, proficiency_demonstration, proficiency_demonstrator, relations, translation, correspondence, atomic_action_descriptions, egopose
""",
    )
    parser.add_argument(
        "--splits",
        type=str,
        nargs="+",
        default=None,
        help="""
Data relating to the train/val/test dataset splits. Valid values are {train, val, test}.
""",
    )
    parser.add_argument(
        "--views",
        type=str,
        nargs="+",
        default=None,
        help="""
Data relating to a particular view. Valid values are {ego, exo}.
""",
    )
    parser.add_argument(
        "-u",
        "--universities",
        type=str,
        nargs="+",
        default=None,
        help=f"""
Data relating to a particular university. Valid values are:
{", ".join(UNIV_TO_BUCKET.keys())}
""",
    )
    parser.add_argument(
        "--num_workers",
        type=int,
        default=15,
        help="number of workers to perform download ops",
    )
    parser.add_argument(
        "-y",
        "--yes",
        default=False,
        help="don't prompt to confirm",
        action="store_true",
    )
    parser.add_argument(
        "-d",
        "--delete",
        default=False,
        help="delete unused files in the directory",
        action="store_true",
    )
    parser.add_argument(
        "--force",
        default=False,
        help="force a download of all files",
        action="store_true",
    )
    parser.add_argument(
        "--release",
        type=str,
        default=release_name,
        help="name/version of the release",
    )
    parser.add_argument(
        "--s3_profile",
        type=str,
        default="default",
        help="profile to use for S3",
    )
    parser.add_argument(
        "--base_dir",
        type=str,
        default=base_dir,
        help="base directory for download (ADVANCED usage, do not change unless you know what you're doing)",
    )
    return parser


def internal_main():
    parser = create_arg_parse(
        script_name="egoexo_internal",
        release_name=None,
        base_dir="s3://ego4d-consortium-sharing/egoexo/releases/",
    )
    args = parser.parse_args()
    main(args)


if __name__ == "__main__":
    internal_main()



import os
import random
import time
import traceback
from typing import Any, List, Optional, Tuple

import boto3
import boto3.session
import botocore.client as bclient
from iopath import PathManager

S3Client = bclient.BaseClient


def _get_location(bucket_name: str) -> str:
    client = boto3.client("s3")
    response = client.get_bucket_location(Bucket=bucket_name)
    return response["LocationConstraint"]


def get_config(
    num_workers: int = 10,
    connect_timeout: int = 180,
    max_attempts: int = 3,
    region_name: Optional[str] = None,
):
    config = bclient.Config(
        region_name=region_name,
        connect_timeout=connect_timeout,
        max_pool_connections=num_workers,
        retries={"mode": "standard", "max_attempts": max_attempts},
    )
    return config


def get_session(profile: Optional[str]):
    return boto3.session.Session(profile_name=profile)


def get_client(
    bucket_name: str,
    num_workers: int = 10,
    connect_timeout: int = 180,
    max_attempts: int = 3,
    profile: Optional[str] = None,
) -> S3Client:
    session = get_session(
        profile=profile,
    )
    return session.client(
        "s3",
        config=get_config(
            region_name=_get_location(bucket_name),
            num_workers=num_workers,
            connect_timeout=connect_timeout,
            max_attempts=max_attempts,
        ),
    )


def get_resource(
    profile: str,
    num_workers: int = 10,
    connect_timeout: int = 180,
    max_attempts: int = 3,
) -> Any:
    session = get_session(profile=profile)
    return session.resource(
        "s3",
        config=get_config(
            region_name=None,
            num_workers=num_workers,
            connect_timeout=connect_timeout,
            max_attempts=max_attempts,
        ),
    )


class StreamPathMgr:
    def __init__(self, expiration_sec: int = 7200):
        self.clients = {}
        self.expr_sec = expiration_sec
        self.cached_paths = {}

    def open(self, path: str) -> str:
        if path in self.cached_paths:
            return self.cached_paths[path]

        if path.startswith("s3"):
            temp = path.split("s3://")[1].split("/")
            bucket_name = temp[0]
            object_name = "/".join(temp[1:])

            client = None
            if bucket_name not in self.clients:
                self.clients[bucket_name] = get_client(bucket_name)

            client = self.clients[bucket_name]
            ret = client.generate_presigned_url(
                "get_object",
                Params={"Bucket": bucket_name, "Key": object_name},
                ExpiresIn=self.expr_sec,
            )
            self.cached_paths[path] = ret
            return ret
        else:
            return path


def ls_relative(path: str, pathmgr: PathManager) -> List[str]:
    files = pathmgr.ls(path)
    return [f.split("/")[-1] for f in files]


# TOOD(miguelmartin): removeme, duplicate function
def exp_backoff(max_sleep_time_sec=1200, base=2):  # pyre-ignore
    def decorator(func):  # pyre-ignore
        assert base > 1

        def wrapper(*args, **kwargs):  # pyre-ignore
            sleep_t = base
            while True:
                try:
                    ret = func(*args, **kwargs)
                    return ret
                except Exception as e:
                    print(
                        f"WARN: exp_backoff retrying...: {func.__name__}, {traceback.format_exc()}"
                    )
                    time.sleep(sleep_t / base + sleep_t * random.random())
                    sleep_t *= base
                    if sleep_t > max_sleep_time_sec:
                        print(
                            f"ERROR: exp_backoff FAILED: {func.__name__}, {traceback.format_exc()}"
                        )
                        raise e
                    continue

        return wrapper

    return decorator


# NOTE:
# iopath / pathmgr not used as we cannot get an S3 Object for sizing with it
class S3Downloader:
    def __init__(self, profile, num_workers=10, callback=None):
        self.clients = {}
        self.resources = get_resource(profile=profile, num_workers=num_workers)
        self.profile = profile
        self.callback = callback

    def ls(self, path, recursive=False, max_keys=-1) -> List[Tuple[str, str]]:
        assert path.startswith("s3://")
        temp = path.split("s3://")[1].split("/")
        bucket_name = temp[0]
        object_name = "/".join(temp[1:])
        if bucket_name not in self.clients:
            self.clients[bucket_name] = get_client(
                bucket_name,
                profile=self.profile,
            )  # pyre-ignore

        delim = "/" if not recursive else ""
        client = self.clients[bucket_name]
        if max_keys < 0:
            paginator = client.get_paginator("list_objects_v2")
            pages = paginator.paginate(
                Bucket=bucket_name, Prefix=object_name, Delimiter=delim
            )
            return [
                (os.path.basename(f["Key"]), f"s3://{bucket_name}/{f['Key']}")
                for page in pages
                for f in page.get("Contents", [])
            ]
        else:
            assert max_keys <= 1000
            ls_result = client.list_objects_v2(
                Bucket=bucket_name,
                Prefix=object_name,
                MaxKeys=max_keys,
                Delimiter="/",
            )
            if ls_result["KeyCount"] == 0:
                return []
            return [
                (os.path.basename(f["Key"]), f"s3://{bucket_name}/{f['Key']}")
                for f in ls_result["Contents"]
            ]

    def copy(self, path: str, out_path: str):
        self.obj(path).download_file(out_path, Callback=self.callback)

    def obj(self, path):
        assert path.startswith("s3://")
        temp = path.split("s3://")[1].split("/")
        bucket_name = temp[0]
        object_name = "/".join(temp[1:])

        return self.resources.Object(bucket_name, object_name)
